// PerM.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "ColorSpaceRead.h"

int main(int argc, char* argv[])
{
    //testShift64Bit();
    testLongBases2Colors();
    return 0;
}

